"""

`foggy_training.models` provides some default implementations for common models. They inherit from LightningExtended
and are trainable with PytorchLightning.

## Example

Say you have your regression module ready:
```python
model = nn.Sequential(...modules)
```
now you want to train it. Use it as follows:
```python
model = ModuleFromModel(model, 'mse_loss', metrics=['R2Score'])
```
and fit:

```python
trainer = Trainer(**config.trainer.as_dict)
trainer.fit(model, ...your data)
```


"""

from functools import singledispatch
from typing import List, Union

import torch
from datasets import Dataset, DatasetDict
from torch import nn
from torch.nn.functional import cross_entropy
from transformers import AutoConfig, AutoModel, AutoTokenizer

from foggy_training.base import LightningExtended, ForwardResult


@singledispatch
def prepare_inputs(dataset: Dataset, x_cols: List[str], y_col: str):
    assert all([f in dataset.features for f in x_cols]), 'x_cols provided that are not in dataset'
    assert y_col in dataset.features, 'y_col provided that is not in dataset'
    remove = [c for c in dataset.features if c not in x_cols + [y_col]]
    dataset = dataset.remove_columns(remove)
    dataset.set_format('pt')
    return dataset


@prepare_inputs.register
def _(dataset: DatasetDict, x_cols: List[str], y_col: str):
    new_data = dict()
    for d in dataset:
        new_data[d] = prepare_inputs(dataset[d], x_cols, y_col)
    return DatasetDict(new_data)


class SoftmaxClassifier(LightningExtended):
    def __init__(self, n_inputs: int, n_classes: int, metrics=None, hidden_dim: int = 50):
        super().__init__(loss=cross_entropy, metrics=metrics)

        self.model = nn.Sequential(
            nn.Linear(n_inputs, hidden_dim), nn.LeakyReLU(),
            nn.Linear(hidden_dim, hidden_dim), nn.LeakyReLU(),
            nn.Linear(hidden_dim, n_classes)
        )

    def forward(self, batch: dict) -> ForwardResult:
        pred = self.model(batch['x'])
        return ForwardResult(pred=pred, true=batch['y'])


class ModuleFromModel(LightningExtended):
    def __init__(self, model, loss: Union[str, callable], lr: float = None, metrics=None, x: str = 'x', y: str = 'y'):
        """
        ModuleFromModel turns an existing model into a LightningExtended Module. This way, it is trainable using
        PytorchLightning.

        Example:
        ```python
        from foggy_training.models import ModuleFromModel
        from foggy_training import config

        model = nn.Sequential(nn.Linear(10, 15), nn.ReLU(), nn.Linear(15, 15), nn.ReLU(), nn.Linear(15, 1))
        model = ModuleFromModel(model, 'mse_loss', metrics=['R2Score'])

        trainer = Trainer(**config.trainer.as_dict)
        trainer.fit(model, ...your data)

        ```

        Args:
            model: nn.Module
            loss: str or callable, loss to use for training
            metrics: str or torchmetrics, metrics to collect during training. defaults to None.
            x: the data is expected to be in a dictionary. x specifies the entry for x values
            y: y specifies the entry for targets
        """
        super().__init__(loss, lr=lr, metrics=metrics)
        self.model = model
        self.x = x
        self.y = y

    def forward(self, batch) -> ForwardResult:
        x, y = batch[self.x], batch.get(self.y, None)
        pred = self.model(x)
        return ForwardResult(pred, y)


class TransformerModel(LightningExtended):
    def __init__(self, backbone_name: str, head_fn: callable, loss: callable, metrics=None):
        super().__init__(loss=loss, metrics=metrics)
        config = AutoConfig.from_pretrained(backbone_name)
        self.backbone_name = backbone_name
        self.backbone = AutoModel.from_pretrained(backbone_name, config=config)
        self.head = head_fn()

    def forward(self, batch: dict) -> ForwardResult:
        batch = self.backbone(input_ids=batch["input_ids"], attention_mask=batch.get("attention_mask", None))
        batch = batch['last_hidden_state']
        batch = self.head(batch)
        return ForwardResult(pred=batch)

    def tokenize(self, dataset, columns, padding='max_length', truncation=True):
        tokenizer = AutoTokenizer.from_pretrained(self.backbone_name)
        dataset = dataset.map(
            lambda examples: tokenizer(examples[columns],
                                       padding=padding, truncation=truncation, return_tensors="np"),
            batched=True, desc='tokenizing'
        )
        return dataset

    def summary(self, depth: int = 3,
                col_names=('output_size', 'num_params', 'trainable'),
                row_settings=('var_names', ), **kwargs):
        """
        Prints summary of the model with chosen depth of child modules. **kwargs are forwarded to torchsummary.
        """
        input_shape = (3, self.backbone.config.max_position_embeddings)
        input_data = dict(input_ids=torch.ones(input_shape).int())

        super(TransformerModel, self).summary(
            depth=depth, input_data=[input_data],
            col_names=col_names, row_settings=row_settings, **kwargs)



from sklearn.base import BaseEstimator, ClassifierMixin


class ClassificationWrapper(BaseEstimator, ClassifierMixin):

    def __init__(self, model_fun: callable, trainer_fun: callable):
        self.model_fun = model_fun
        self.trainer_fun = trainer_fun

    def fit(self, X, y, X_val=None, y_val=None, **kwargs):
        self.model_ = self.model_fun()
        self.trainer_ = self.trainer_fun()
        inputs = prepare_inputs(dataset, ['x'], 'y')
        train_dl = config.data.get_dataloader(inputs['train'], shuffle=True)
        val_dl = config.data.get_dataloader(inputs['test'], shuffle=False)







"""

The config bundles the configuration for dataset, model and training.
```python
from foggy_training import config

print(config)
# config was set to cpu because no gpu was detected.
# Config(trainer=TrainerConfig(max_epochs=3, ...), data=DataConfig(batch_size=32, num_workers=0))
```

replace values:

```python
config.trainer.max_epochs = 15
```

get trainer:

```python
from pytorch_lightning import Trainer
trainer = Trainer(**config.trainer.as_dict)

```


"""



from typing import Union, List

import torch
from attr import field, asdict, define
from datasets import Dataset
from pytorch_lightning import Callback
from torch.utils.data import DataLoader


@define
class DataConfig:
    """

    DataConfig holds values that are specific for dataset creation:

    batch size (defaults to 32)
    num_workers (defaults to 6 on GPU and 0 on CPU)

    """
    batch_size: int = 32
    num_workers: int = 6

    def cpu(self):
        self.num_workers = 0
        self.batch_size = 32

    def gpu(self):
        self.num_workers = 6
        self.batch_size = 32

    @property
    def as_dict(self):
        return asdict(self)  # noqa

    def get_dataloader(self, dataset: Dataset, shuffle: bool) -> DataLoader:
        """
        The config may create DataLoaders based on the config values

        Args:
            dataset: torch compatible dataset
            shuffle: if True, shuffle values.

        Returns: DataLoader

        """
        return DataLoader(dataset, shuffle=shuffle, **self.as_dict)


@define
class TrainerConfig:

    """

    TrainerConfig holds values that are specific for trainer creation:

    max_epochs (defaults to 20)
    ... tbc

    """

    max_epochs: int = 20
    auto_scale_batch_size: Union[str, bool] = False
    callbacks: List[Callback] = field(factory=lambda: [])
    accelerator: str = 'gpu'
    precision: Union[int, str] = 16
    logger: bool = True
    gradient_clip_val: Union[int, float] = None
    enable_progress_bar: bool = True
    overfit_batches: Union[int, float] = 0.0
    track_grad_norm: Union[int, float, str] = -1
    fast_dev_run: Union[int, bool] = False
    log_every_n_steps: int = 50
    enable_model_summary: bool = True
    num_sanity_val_steps: int = 2

    def cpu(self):
        self.accelerator = 'cpu'
        self.precision = 32
        self.logger = False
        self.max_epochs = 3

    def gpu(self):
        self.accelerator = 'gpu'
        self.precision = 16
        self.logger = True
        self.overfit_batches = 0.0

    @property
    def as_dict(self):
        return asdict(self)  # noqa


@define
class Config:
    """

    Config holds the values for trainer and data config.

    You can change default values by calling config.cpu() or config.gpu().

    """
    trainer: TrainerConfig = TrainerConfig()
    data: DataConfig = DataConfig()

    def __attrs_post_init__(self):
        if not torch.cuda.is_available():
            self.cpu()
            print('config was set to cpu because no gpu was detected.')

    def cpu(self):
        self.trainer.cpu()
        self.data.cpu()

    def gpu(self):
        self.trainer.gpu()
        self.data.gpu()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['foggy_training',
 'foggy_training.addons',
 'foggy_training.base',
 'foggy_training.datasets',
 'foggy_training.models',
 'foggy_training.wrapper']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.1.0,<23.0.0', 'torchinfo>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'foggy-training',
    'version': '0.1.154',
    'description': '',
    'long_description': None,
    'author': 'JanDiers',
    'author_email': 'jan.diers@uni-jena.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Union

import pytorch_lightning as pl
from attr import define, field
from torch import optim
from torchinfo import summary
from torchmetrics import MetricCollection

from foggy_training.base.lookups import get_loss_fun, get_metric_fun


@define
class ForwardResult:
    pred = field()
    true = field(default=None)
    loss = field(default=None)


class LightningExtended(pl.LightningModule):

    """

    The Lightning Extended Modules provides some defaults for training and validation sets. You may do the following:

    ```python
    from foggy_training import models
    x = [[1., 1.], [2., 2.], [3., 3.]]
    y = [1., 2., 3.]
    data = {'x': x, 'y': y}
    model = models.LinearRegression(n_inputs=2)
    model.fit(data, epochs=20)
    ```

    """
    def __init__(self, loss: Union[str, callable], lr: float = None, metrics=None):
        """

        loss: string or callable. if string, must match the name of torch.nn.functional losses
        lr: float, learning rate
        metrics: list of torchmetrics

        """
        super(LightningExtended, self).__init__()
        self.loss = get_loss_fun(loss)
        self.lr = lr if lr is not None else 1e-3
        metrics = [] if metrics is None else metrics
        assert isinstance(metrics, list), 'metrics must be list of metrics'
        metrics = [get_metric_fun(m) for m in metrics]
        metrics = MetricCollection(metrics)
        self.train_metrics = metrics.clone(prefix='train_')
        self.val_metrics = metrics.clone(prefix='val_')

    def shared_step(self, batch) -> ForwardResult:
        result = self(batch)
        return result

    def training_step(self, batch, batch_idx):
        result = self.shared_step(batch)
        loss = self.loss(result.pred, result.true)
        logs = dict(loss=loss)
        metrics = self.train_metrics(result.pred, result.true)
        logs = {**logs, **metrics}
        self.log_dict(logs, prog_bar=True, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        result = self.shared_step(batch)
        loss = self.loss(result.pred, result.true)
        logs = dict(val_loss=loss)
        metrics = self.val_metrics(result.pred, result.true)
        logs = {**logs, **metrics}
        self.log_dict(logs, prog_bar=True, on_step=False, on_epoch=True)

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        return self.shared_step(batch).pred

    def configure_optimizers(self):
        return optim.AdamW(filter(lambda p: p.requires_grad, self.parameters()), lr=self.lr)

    def summary(self, depth: int = 3,
                col_names=('num_params', 'trainable'),
                row_settings=('var_names', ), **kwargs):
        print(summary(self, depth=depth,
                      col_names=col_names,
                      row_settings=row_settings,
                      **kwargs))
